from dynabuffers.api.ISerializable import ISerializable


class BooleanType(ISerializable):

    def size(self, value, registry):
        return 1

    def serialize(self, value, buffer, registry):
        if value in [True, "true"]:
            buffer.put(bytes([1]))
        else:
            buffer.put(bytes([0]))

    def deserialize(self, buffer, registry):
        if buffer.get() == 0:
            return False
        else:
            return True



# Generated from Dynabuffers.g4 by ANTLR 4.7
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .DynabuffersParser import DynabuffersParser
else:
    from DynabuffersParser import DynabuffersParser

# This class defines a complete listener for a parse tree produced by DynabuffersParser.
class DynabuffersListener(ParseTreeListener):

    # Enter a parse tree produced by DynabuffersParser#compilation.
    def enterCompilation(self, ctx:DynabuffersParser.CompilationContext):
        pass

    # Exit a parse tree produced by DynabuffersParser#compilation.
    def exitCompilation(self, ctx:DynabuffersParser.CompilationContext):
        pass


    # Enter a parse tree produced by DynabuffersParser#enumType.
    def enterEnumType(self, ctx:DynabuffersParser.EnumTypeContext):
        pass

    # Exit a parse tree produced by DynabuffersParser#enumType.
    def exitEnumType(self, ctx:DynabuffersParser.EnumTypeContext):
        pass


    # Enter a parse tree produced by DynabuffersParser#classType.
    def enterClassType(self, ctx:DynabuffersParser.ClassTypeContext):
        pass

    # Exit a parse tree produced by DynabuffersParser#classType.
    def exitClassType(self, ctx:DynabuffersParser.ClassTypeContext):
        pass


    # Enter a parse tree produced by DynabuffersParser#unionType.
    def enterUnionType(self, ctx:DynabuffersParser.UnionTypeContext):
        pass

    # Exit a parse tree produced by DynabuffersParser#unionType.
    def exitUnionType(self, ctx:DynabuffersParser.UnionTypeContext):
        pass


    # Enter a parse tree produced by DynabuffersParser#fieldType.
    def enterFieldType(self, ctx:DynabuffersParser.FieldTypeContext):
        pass

    # Exit a parse tree produced by DynabuffersParser#fieldType.
    def exitFieldType(self, ctx:DynabuffersParser.FieldTypeContext):
        pass


    # Enter a parse tree produced by DynabuffersParser#dataType.
    def enterDataType(self, ctx:DynabuffersParser.DataTypeContext):
        pass

    # Exit a parse tree produced by DynabuffersParser#dataType.
    def exitDataType(self, ctx:DynabuffersParser.DataTypeContext):
        pass


    # Enter a parse tree produced by DynabuffersParser#arrayType.
    def enterArrayType(self, ctx:DynabuffersParser.ArrayTypeContext):
        pass

    # Exit a parse tree produced by DynabuffersParser#arrayType.
    def exitArrayType(self, ctx:DynabuffersParser.ArrayTypeContext):
        pass


    # Enter a parse tree produced by DynabuffersParser#optionType.
    def enterOptionType(self, ctx:DynabuffersParser.OptionTypeContext):
        pass

    # Exit a parse tree produced by DynabuffersParser#optionType.
    def exitOptionType(self, ctx:DynabuffersParser.OptionTypeContext):
        pass


    # Enter a parse tree produced by DynabuffersParser#namespaceType.
    def enterNamespaceType(self, ctx:DynabuffersParser.NamespaceTypeContext):
        pass

    # Exit a parse tree produced by DynabuffersParser#namespaceType.
    def exitNamespaceType(self, ctx:DynabuffersParser.NamespaceTypeContext):
        pass


    # Enter a parse tree produced by DynabuffersParser#classOptions.
    def enterClassOptions(self, ctx:DynabuffersParser.ClassOptionsContext):
        pass

    # Exit a parse tree produced by DynabuffersParser#classOptions.
    def exitClassOptions(self, ctx:DynabuffersParser.ClassOptionsContext):
        pass


    # Enter a parse tree produced by DynabuffersParser#unionOptions.
    def enterUnionOptions(self, ctx:DynabuffersParser.UnionOptionsContext):
        pass

    # Exit a parse tree produced by DynabuffersParser#unionOptions.
    def exitUnionOptions(self, ctx:DynabuffersParser.UnionOptionsContext):
        pass


    # Enter a parse tree produced by DynabuffersParser#fieldOptions.
    def enterFieldOptions(self, ctx:DynabuffersParser.FieldOptionsContext):
        pass

    # Exit a parse tree produced by DynabuffersParser#fieldOptions.
    def exitFieldOptions(self, ctx:DynabuffersParser.FieldOptionsContext):
        pass


    # Enter a parse tree produced by DynabuffersParser#annotation.
    def enterAnnotation(self, ctx:DynabuffersParser.AnnotationContext):
        pass

    # Exit a parse tree produced by DynabuffersParser#annotation.
    def exitAnnotation(self, ctx:DynabuffersParser.AnnotationContext):
        pass


    # Enter a parse tree produced by DynabuffersParser#value.
    def enterValue(self, ctx:DynabuffersParser.ValueContext):
        pass

    # Exit a parse tree produced by DynabuffersParser#value.
    def exitValue(self, ctx:DynabuffersParser.ValueContext):
        pass





import struct
from abc import ABC, abstractmethod
from struct import unpack


class ISerializable(ABC):

    @abstractmethod
    def size(self, value, registry):
        pass

    @abstractmethod
    def serialize(self, value, buffer: 'ByteBuffer', registry):
        pass

    @abstractmethod
    def deserialize(self, buffer: 'ByteBuffer', registry):
        pass

    def byte(self, obj):
        return bytes(obj)[0]


class ByteBuffer():

    def __init__(self, length:int, buffer:bytearray = bytearray()):
        self.length = length
        self.buffer = buffer

    @staticmethod
    def wrap(buffer: bytearray) -> 'ByteBuffer':
        return ByteBuffer(len(buffer), buffer)

    def put(self, obj: bytes):
        self.buffer = self.buffer + obj

    def putShort(self, obj:int):
        self.buffer = self.buffer + struct.pack('>h', obj)

    def putInt(self, obj:int):
        self.buffer = self.buffer + struct.pack('>i', obj)

    def putFloat(self, obj:float):
        self.buffer = self.buffer + struct.pack('>f', obj)

    def putLong(self, obj:int):
        self.buffer = self.buffer + struct.pack('>q', obj)

    def has_remaining(self):
        return len(self.buffer) > 0

    def get(self, length=None):
        if length is None:
            return self.buffer.pop(0)
        else:
            bytes = self.buffer[:length]
            self.buffer = self.buffer[length:]
            return bytes

    def getShort(self):
        buffer = self.buffer[:2]
        self.buffer = self.buffer[2:]
        # https://stackoverflow.com/questions/45187101/converting-bytearray-to-short-int-in-python
        return unpack('>h', buffer)[0]

    def getInt(self):
        buffer = self.buffer[:4]
        self.buffer = self.buffer[4:]
        return unpack('>i', buffer)[0]

    def getFloat(self):
        bytes = self.buffer[:4]
        self.buffer = self.buffer[4:]
        return struct.unpack('>f', bytes)[0]

    def getLong(self):
        bytes = self.buffer[:8]
        self.buffer = self.buffer[8:]
        return struct.unpack('>q', bytes)[0]

    def toBytes(self):
        return self.buffer



from dynabuffers.api.IAnnotation import IAnnotation
from dynabuffers.ast.structural import Value


class LowerEquals(IAnnotation):

    def __init__(self, args:[Value]):
        self.args = args
        if len(args) != 1:
            raise ValueError("too much annotation values")
        if not str(args[0].options.value).isnumeric():
            raise ValueError(str(args[0].options.value) + " is not a number")

    def validate(self, fieldName, value):
        if float(value) > float(self.args[0].options.value):
            raise ValueError("(" + str(fieldName) + ":" + str(value) + ") is greater than " + str(self.args[0].options.value))



from dynabuffers.api.ISerializable import ISerializable
from dynabuffers.ast.datatype.ByteType import ByteType


class ArrayTypeOptions:
    def __init__(self, datatype:ISerializable):
        self.datatype = datatype

class ArrayType(ISerializable):

    def __init__(self, options):
        self.options = options

    def size(self, value, registry):
        s = sum([self.options.datatype.size(item, registry) for item in value])
        return 4 + s

    def serialize(self, value, buffer, registry):
        buffer.putInt(len(value))
        for entry in value:
            self.options.datatype.serialize(entry, buffer, registry)

    def deserialize(self, buffer, registry):
        length = buffer.getInt()
        array = []

        for i in range(length):
            array.append(self.options.datatype.deserialize(buffer, registry))

        if isinstance(self.options.datatype, ByteType):
            return bytes(array)
        return array



from dynabuffers.api.ISerializable import ISerializable


class RefTypeOptions:
    def __init__(self, name: str):
        self.name = name


class RefType(ISerializable):
    def __init__(self, options: RefTypeOptions):
        self.options = options

    def size(self, value, registry):
        return registry.resolve(self.options.name).size(value, registry)

    def serialize(self, value, buffer, registry):
        registry.resolve(self.options.name).serialize(value, buffer, registry)

    def deserialize(self, buffer, registry):
        return registry.resolve(self.options.name).deserialize(buffer, registry)



from dynabuffers.api.ISerializable import ISerializable
from dynabuffers.ast.ClassType import ClassType
from dynabuffers.ast.EnumType import EnumType
from dynabuffers.ast.UnionType import UnionType
from dynabuffers.ast.annotation.GreaterEquals import GreaterEquals
from dynabuffers.ast.annotation.GreaterThan import GreaterThan
from dynabuffers.ast.annotation.LowerEquals import LowerEquals
from dynabuffers.ast.annotation.LowerThan import LowerThan
from dynabuffers.ast.annotation.MaxLength import MaxLength
from dynabuffers.ast.annotation.MinLength import MinLength
from dynabuffers.ast.annotation.NotBlank import NotBlank


class Registry(object):

    def __init__(self, tree: [ISerializable], listeners):
        self.tree = tree
        self.listeners = listeners

    def resolveAnnotation(self, name, args):
        if name == "GreaterThan":
            return GreaterThan(args)
        if name == "GreaterEquals":
            return GreaterEquals(args)
        if name == "LowerThan":
            return LowerThan(args)
        if name == "LowerEquals":
            return LowerEquals(args)
        if name == "MaxLength":
            return MaxLength(args)
        if name == "MinLength":
            return MinLength(args)
        if name == "NotBlank":
            return NotBlank(args)
        raise ValueError("unknown annotation " + str(name))

    def createForSchema(self, schema: [ISerializable]):
        return Registry(schema, self.listeners)

    def resolve(self, name: str):
        for element in self.tree:
            if isinstance(element, ClassType) and element.options.name == name:
                return element
            if isinstance(element, EnumType) and element.options.name == name:
                return element
            if isinstance(element, UnionType) and element.options.name == name:
                return element

    def addNotification(self, notification: str):
        for listener in self.listeners:
            listener(notification)



from typing import Union, Any, List

from dynabuffers import NAMESPACE_KEY
from dynabuffers.NamespaceResolver import NamespaceResolver
from dynabuffers.Registry import Registry
from dynabuffers.api.ISerializable import ISerializable, ByteBuffer
from dynabuffers.api.map.DynabuffersMap import DynabuffersMap
from dynabuffers.header.RootElement import RootElement


class DynabuffersEngine(object):

    def __init__(self, tree: [ISerializable]):
        self.tree = tree
        self.namespace_resolver = NamespaceResolver(tree)
        self.listeners = []

    def add_listener(self, listener):
        self.listeners.append(listener)

    def serialize(self, value: Union[dict, Any], namespace_names: Union[List[str], str] = None) -> bytearray:
        if not isinstance(value, dict):
            return self.serialize({"value": value})

        if namespace_names is not None:
            if isinstance(namespace_names, List):
                value[NAMESPACE_KEY] = '.'.join(namespace_names)
            if isinstance(namespace_names, str):
                value[NAMESPACE_KEY] = namespace_names

        root_element = RootElement(self.tree)
        registry = Registry(self.tree, self.listeners)
        buffer = ByteBuffer(root_element.size(value, registry))
        root_element.serialize(value, buffer, registry)
        return buffer.toBytes()

    def deserialize(self, bytes: bytearray, namespaces: Union[List[str], str] = None) -> DynabuffersMap:
        """
        :param bytes:
        :param namespaces: only there for compat reasons => unused and deprecated!
        :return:
        """
        bb = ByteBuffer(len(bytes), bytes)
        registry = Registry(self.tree, self.listeners)
        return RootElement(self.tree).deserialize(bb, registry)



from typing import List, Union

from dynabuffers import NAMESPACE_KEY
from dynabuffers.DynabuffersEngine import Registry
from dynabuffers.header.DynabuffersVersion import DYNABUFFERS_VERSION
from dynabuffers.header.Header import Header
from dynabuffers.NamespaceResolver import NamespaceResolver, NamespaceDescription
from dynabuffers.api.ISerializable import ISerializable
from dynabuffers.api.map.DynabuffersMap import DynabuffersMap
from dynabuffers.api.map.ImplicitDynabuffersMap import ImplicitDynabuffersMap
from dynabuffers.ast.ClassType import ClassType
from dynabuffers.ast.UnionType import UnionType


class RootElement(ISerializable):
    def __init__(self, tree: [ISerializable], version: int = DYNABUFFERS_VERSION.get_major()):
        self._tree = tree
        self._version = version
        self._namespace_resolver = NamespaceResolver(tree)

    def size(self, value, registry):
        schema = self.__get_schema_for_namespace(value)
        header_size = Header(self._namespace_resolver, self._version).size(value, registry)
        content_size = self.__get_root_type(schema).size(value, registry.createForSchema(schema))
        return header_size + content_size

    def __get_schema_for_namespace(self, value: dict) -> List[ISerializable]:
        namespace_description = self._namespace_resolver.get_namespace(value)
        if namespace_description is not None:
            return namespace_description.namespace.options.list
        return self._tree

    def serialize(self, value, buffer: 'ByteBuffer', registry):
        header = Header(self._namespace_resolver, self._version)
        header.serialize(value, buffer, registry)
        schema = self.__get_schema_for_namespace(value)
        self.__get_root_type(schema).serialize(value, buffer, registry.createForSchema(schema))

    def deserialize(self, buffer: 'ByteBuffer', registry):
        header = Header(self._namespace_resolver, self._version).deserialize(buffer, registry)
        if header.version != self._version:
            raise ValueError("Dynabuffers version of serialized data does not match current version. "
                             "Expected: {}  Actual: {}".format(self._version, header.version))

        # TODO: Handle flag bits here

        if header.namespace_description is not None:
            schema = header.namespace_description.namespace.options.list
            new_root = RootElement(schema)
            new_registry = registry.createForSchema(schema)
            return new_root.__deserialize_content(buffer, new_registry, header.namespace_description)
        return self.__deserialize_content(buffer, registry, None)

    def __deserialize_content(self, buffer: 'ByteBuffer', registry,
                              namespace_description: Union[NamespaceDescription, None]):
        root_type = self.__get_root_type(self._tree)
        result = root_type.deserialize(buffer, registry)
        if namespace_description is not None:
            result[NAMESPACE_KEY] = namespace_description.joined_path()

        if root_type.options.options.is_implicit:
            return ImplicitDynabuffersMap(result, self._tree, root_type)
        return DynabuffersMap(result, self._tree, root_type)

    def __get_root_type(self, schema: [ISerializable]) -> ClassType:
        classes = list(filter(lambda x: isinstance(x, ClassType) or isinstance(x, UnionType), schema))
        for clazz in classes:
            if clazz.options.options.is_primary():
                return clazz

        if len(classes) == 0:
            raise ValueError("no root type found")
        return classes[0]



from dynabuffers.api.ISerializable import ISerializable, ByteBuffer
from dynabuffers.ast import FieldType
from dynabuffers.ast.datatype.OptionType import OptionType
from dynabuffers.ast.structural import ClassOptions


class ClassTypeOptions:
    def __init__(self, name, fields: [FieldType], options: ClassOptions):
        self.name = name
        self.fields = fields
        self.options = options


class ClassType(ISerializable):

    def __init__(self, options: ClassTypeOptions):
        self.options = options

    def size(self, value, registry):
        for field in self.options.fields:
            if field.options.defaultVal is None and field.options.name not in value:
                if not isinstance(field.options.datatype, OptionType):
                    raise ValueError("field '" + str(field.options.name) + "' is missing")
                else:
                    value[field.options.name] = None
            if field.options.defaultVal is None and value[field.options.name] is None:
                if not isinstance(field.options.datatype, OptionType):
                    raise ValueError("field '" + str(field.options.name) + "' is missing")
                else:
                    value[field.options.name] = None

        size = 0
        for field in self.options.fields:
            if field.options.name in value:
                size += field.size(value[field.options.name], registry)
            elif field.options.defaultVal is not None:
                size += field.size(field.options.defaultVal, registry)
        return size

    def serialize(self, value, buffer: ByteBuffer, registry):
        if self.options.options.is_deprecated():
            registry.addNotification("deprecated class " + self.options.name + " used")

        for field in self.options.fields:
            if not isinstance(field.options.datatype, OptionType):
                if field.options.defaultVal is None and field.options.name not in value:
                    raise ValueError("field '" + str(field.options.name) + "' is missing")
                if field.options.defaultVal is None and value[field.options.name] is None:
                    raise ValueError("field '" + str(field.options.name) + "' is missing")

        for element in list(
                filter(lambda x: (x.options.name in value) or x.options.defaultVal is not None, self.options.fields)):
            element.serialize(
                value[element.options.name] if element.options.name in value else element.options.defaultVal, buffer,
                registry)

    def deserialize(self, buffer, registry):
        map = {}
        for field in self.options.fields:
            map[field.options.name] = field.deserialize(buffer, registry)

        return map



from dynabuffers.api.ISerializable import ISerializable, ByteBuffer
from dynabuffers.ast.datatype.ArrayType import ArrayType, ArrayTypeOptions
from dynabuffers.ast.datatype.BooleanType import BooleanType
from dynabuffers.ast.datatype.ByteType import ByteType
from dynabuffers.ast.datatype.FloatType import FloatType
from dynabuffers.ast.datatype.IntType import IntType
from dynabuffers.ast.datatype.LongType import LongType
from dynabuffers.ast.datatype.NoneType import NoneType
from dynabuffers.ast.datatype.ShortType import ShortType
from dynabuffers.ast.datatype.StringType import StringType, StringTypeOptions


class MapTypeOptions:
    def __init__(self, charset: str):
        self.charset = charset


class MapType(ISerializable):

    def __init__(self, options: MapTypeOptions):
        self.options = options

    def size(self, value, registry):
        size = 2
        for key in value:
            if key is None:
                continue

            key_type = self._get_key_type(key)

            size += 4
            size += key_type.size(key, registry)

            if value[key] is not None:
                val_type = self._get_val_type(value[key])
                size += val_type.size(value[key], registry)

        return size

    def serialize(self, value, buffer: ByteBuffer, registry):
        buffer.putShort(len(value))
        for key in value:
            if key is None:
                continue

            key_type = self._get_key_type(key)
            val_type = self._get_val_type(value[key])

            key_size = key_type.size(key, registry)
            val_size = val_type.size(value[key], registry)

            ordinal = self._type_to_ordinal(value[key])
            buffer.putInt(self.merge_header(ordinal, key_size + val_size))
            key_type.serialize(key, buffer, registry)
            val_type.serialize(value[key], buffer, registry)

    def deserialize(self, buffer: ByteBuffer, registry):
        result = dict()

        counter = 0
        amount = buffer.getShort()
        while counter < amount:
            header = buffer.getInt()
            (type, length) = self.unmerge_header(header)

            key_type = StringType(StringTypeOptions(self.options.charset))
            val_type = self._ordinal_to_type(type)

            raw_value = buffer.get(length)
            entry = ByteBuffer.wrap(raw_value)

            k = key_type.deserialize(entry, registry)
            v = val_type.deserialize(entry, registry)

            result[k] = v
            counter += 1

        return result

    def _get_val_type(self, obj) -> ISerializable:
        if isinstance(obj, str):
            return StringType(StringTypeOptions(self.options.charset))
        if isinstance(obj, dict):
            return self
        if isinstance(obj, bool):
            return BooleanType()
        if isinstance(obj, float):
            return FloatType()
        if isinstance(obj, int):
            return IntType() if -2147483648 <= obj <= 2147483647 else LongType()
        if isinstance(obj, bytes):
            return ByteType()
        if isinstance(obj, list):
            return ArrayType(ArrayTypeOptions(self._get_val_type("" if len(obj) == 0 else obj[0])))
        if obj is None:
            return NoneType()
        raise NameError("cannot handle value " + str(obj))

    def _get_key_type(self, obj) -> ISerializable:
        if isinstance(obj, str):
            return StringType(StringTypeOptions(self.options.charset))
        raise NameError("cannot handle value " + str(obj))

    def _type_to_ordinal(self, obj):
        if isinstance(obj, str):
            return 0
        if isinstance(obj, bool):
            return 10
        if isinstance(obj, bytes) and len(obj) == 1:
            return 20
        if isinstance(obj, float):
            return 30
        if isinstance(obj, int):
            return 40 if -2147483648 <= obj <= 2147483647 else 50
        if isinstance(obj, bytes) and len(obj) == 8:
            return 50
        if isinstance(obj, bytes) and len(obj) == 2:
            return 60
        if isinstance(obj, dict):
            return 70
        if isinstance(obj, list):
            return int(80 + self._type_to_ordinal("" if len(obj) == 0 else obj[0]) / 10)
        if obj is None:
            return 90
        raise NameError("cannot handle value " + str(obj))

    def _ordinal_to_type(self, obj):
        if obj == 0:
            return StringType(StringTypeOptions(self.options.charset))
        if obj == 10:
            return BooleanType()
        if obj == 20:
            return ByteType()
        if obj == 30:
            return FloatType()
        if obj == 40:
            return IntType()
        if obj == 50:
            return LongType()
        if obj == 60:
            return ShortType()
        if obj == 70:
            return self
        if obj == 80:
            return ArrayType(ArrayTypeOptions(StringType(StringTypeOptions(self.options.charset))))
        if obj == 81:
            return ArrayType(ArrayTypeOptions(BooleanType()))
        if obj == 82:
            return ArrayType(ArrayTypeOptions(ByteType()))
        if obj == 83:
            return ArrayType(ArrayTypeOptions(FloatType()))
        if obj == 84:
            return ArrayType(ArrayTypeOptions(IntType()))
        if obj == 85:
            return ArrayType(ArrayTypeOptions(LongType()))
        if obj == 86:
            return ArrayType(ArrayTypeOptions(ShortType()))
        if obj == 87:
            return ArrayType(ArrayTypeOptions(self))
        if obj == 90:
            return NoneType()
        raise NameError("cannot handle value " + str(obj))

    def merge_header(self, a: int, b: int) -> int:
        right = 0xFFFFFF
        return a << 24 | (b & right)

    def unmerge_header(self, c: int) -> (int, int):
        right = 0xFFFFFF
        return c >> 24, c & right



from dynabuffers.api.ISerializable import ISerializable


class ShortType(ISerializable):

    def size(self, value, registry):
        return 2

    def serialize(self, value, buffer, registry):
        buffer.putShort(value if isinstance(value, int) else int(value))

    def deserialize(self, buffer, registry):
        return buffer.getShort()



from typing import List


def assert_only_n_first_bits_set(value: int, n: int, name: str):
    if ((0xFF >> (8 - n)) & value) != value:
        raise ValueError("Value {} of {} is too large for field size of {} bits.".format(value, name, n))
    return value


class Subbyte:

    def __init__(self, value: int, number_of_bits: int, name: str):
        """
        Creates and validates a Subbyte (value smaller or equal to a byte in size) with the given value and
        number_of_bits as size in bits.
        :param value: value of the Subbyte
        :param number_of_bits: the size of the Subbyte in bits
        :param name: name of the Subbyte used for logging errors
        :return: Subbyte
        """
        if number_of_bits > 8 or number_of_bits < 1:
            raise ValueError("Field size of {} bits for field {} must be in range 1-8.".format(number_of_bits, name))
        assert_only_n_first_bits_set(value, number_of_bits, name)
        self._value = value
        self._number_of_bits = number_of_bits

    @property
    def value(self):
        return self._value

    @property
    def number_of_bits(self):
        return self._number_of_bits

    @staticmethod
    def compress_values_into_bytes(values: List['Subbyte']) -> List[bytes]:
        result = []
        subbytes_for_byte = []
        total_bits = 0
        for subbyte in values:
            subbytes_for_byte.append(subbyte)
            total_bits += subbyte.number_of_bits
            if total_bits >= 8:
                result.append(Subbyte.compress_values_into_byte(subbytes_for_byte))
                subbytes_for_byte.clear()
                total_bits = 0
        if total_bits != 0:
            raise ValueError("Provided bits ({}) must be a multiple of 8.".format((len(result) * 8) + total_bits))
        return result

    @staticmethod
    def compress_values_into_byte(values: List['Subbyte']) -> bytes:
        """
        Compresses multiple Subbyte values into a byte value if they have the correct number of bits in total.
        :param values: list of Subbyte values
        :return: bytes
        :raise ValueError if the Subbytes values do not have 8 bits in total.
        """
        remaining_bits = 8
        result = 0
        for value in values:
            result |= value.value << (remaining_bits - value.number_of_bits)
            remaining_bits -= value.number_of_bits
        if remaining_bits != 0:
            raise ValueError("Subbyte values ({}) have invalid length in bits. "
                             "Expected: 8 Actual: {}".format(values, 8 - remaining_bits))
        return bytes([result])



from math import ceil
from typing import List, Union

from dynabuffers.NamespaceResolver import NamespaceDescription, NamespaceResolver
from dynabuffers.api.ISerializable import ISerializable
from dynabuffers.api.Subbyte import Subbyte


class HeaderSpec:
    def __init__(self, version: int, flags: int, namespace_description: NamespaceDescription):
        self._version = version
        self._flags = flags
        self._namespace_description = namespace_description

    @property
    def version(self):
        return self._version

    @property
    def flags(self):
        return self._flags

    @property
    def namespace_description(self):
        return self._namespace_description


class Header(ISerializable):
    _number_of_flag_bits = 5

    def __init__(self, namespace_resolver: NamespaceResolver, version: int, flags: int = 0):
        self._namespace_resolver = namespace_resolver
        self._version = version
        self._flags = flags

    def size(self, value, registry):
        namespace_path_size = len(self.get_namespace_path(value))
        return 2 + ceil(namespace_path_size / 2.0)

    def get_namespace_path(self, value: Union[dict, any]) -> List[int]:
        namespace_description = self._namespace_resolver.get_namespace(value)
        if namespace_description is None:
            return []
        return list(map(lambda waypoint: waypoint.position, namespace_description.path))

    def serialize(self, value, buffer: 'ByteBuffer', registry):
        # 1. Byte: Version
        buffer.put(bytes([self._version]))

        # 2. Byte: Namespace depth + flag bits
        namespace_path = self.get_namespace_path(value)
        namespace_depth = Subbyte(len(namespace_path), 8 - self._number_of_flag_bits, "Namespace Depth")
        flag_bits = Subbyte(self._flags, self._number_of_flag_bits, "Flags")
        buffer.put(Subbyte.compress_values_into_byte([namespace_depth, flag_bits]))

        # 3. - 6. Byte: Namespace path
        for byte in self.compress_4_bit_values_to_bytes(namespace_path):
            buffer.put(byte)

    def deserialize(self, buffer: 'ByteBuffer', registry):
        deserialized_version = buffer.get()

        second_byte = buffer.get()
        flag_bits = second_byte & (0xFF >> (8 - self._number_of_flag_bits))
        namespace_depth = second_byte >> self._number_of_flag_bits

        namespace_description = self._namespace_resolver.get_namespace_from_serialized(buffer, namespace_depth)
        return HeaderSpec(deserialized_version, flag_bits, namespace_description)

    def compress_4_bit_values_to_bytes(self, bits: List[int]) -> List[bytes]:
        bits = bits[:]
        if len(bits) % 2 == 1:
            bits.append(0)
        subbytes = [Subbyte(subbyte, 4, "Namespace Path Indicator") for subbyte in bits]
        return Subbyte.compress_values_into_bytes(subbytes)



# This file is autogenerated from DynabuffersVersion.vm

class VersionWrapper:
    def __init__(self, version: str):
        self._version = version

    def get_major(self) -> int:
        try:
            return int(self._version.split(".")[0])
        except ValueError as err:
            raise ValueError("Dynabuffers major version cannot be extracted from: {}".format(self._version)) from err


DYNABUFFERS_VERSION = VersionWrapper("2.6.1")



from dynabuffers.antlr.DynabuffersParser import DynabuffersParser
from dynabuffers.antlr.DynabuffersVisitor import DynabuffersVisitor as DynabuffersBaseVisitor
from dynabuffers.ast.ClassType import ClassType, ClassTypeOptions
from dynabuffers.ast.EnumType import EnumType, EnumTypeOptions
from dynabuffers.ast.FieldType import FieldType, FieldTypeOptions
from dynabuffers.ast.NamespaceType import NamespaceType, NamespaceTypeOptions
from dynabuffers.ast.UnionType import UnionType, UnionTypeOptions
from dynabuffers.ast.datatype.ArrayType import ArrayType, ArrayTypeOptions
from dynabuffers.ast.datatype.BooleanType import BooleanType
from dynabuffers.ast.datatype.ByteType import ByteType
from dynabuffers.ast.datatype.FloatType import FloatType
from dynabuffers.ast.datatype.IntType import IntType
from dynabuffers.ast.datatype.LongType import LongType
from dynabuffers.ast.datatype.MapType import MapType, MapTypeOptions
from dynabuffers.ast.datatype.OptionType import OptionType, OptionTypeOptions
from dynabuffers.ast.datatype.RefType import RefType, RefTypeOptions
from dynabuffers.ast.datatype.ShortType import ShortType
from dynabuffers.ast.datatype.StringType import StringType, StringTypeOptions
from dynabuffers.ast.structural.Annotation import AnnotationOptions, Annotation
from dynabuffers.ast.structural.ClassOptions import ClassOptions, ClassOptionsOptions
from dynabuffers.ast.structural.FieldOptions import FieldOptions, FieldOptionsOptions
from dynabuffers.ast.structural.UnionOptions import UnionOptions, UnionOptionsOptions
from dynabuffers.ast.structural.Value import Value, ValueOptions
from dynabuffers.TypeValidator import DynabuffersTypeValidator


class DynabuffersVisitor(DynabuffersBaseVisitor):

    def __init__(self):
        self.type_validator = DynabuffersTypeValidator()

    def visitCompilation(self, ctx: DynabuffersParser.CompilationContext):
        result = self.visitChildren(ctx)
        self.type_validator.validate()
        return result

    def visitEnumType(self, ctx: DynabuffersParser.EnumTypeContext):
        self.type_validator.add_available_type(ctx)
        name = str(ctx.getChild(1))
        values = []
        for index in range(3, ctx.getChildCount() - 1):
            values.append(str(ctx.getChild(index)))
        return EnumType(EnumTypeOptions(name, values))

    def visitClassType(self, ctx: DynabuffersParser.ClassTypeContext):
        self.type_validator.add_available_type(ctx)
        values = super().visitClassType(ctx)
        options = next(filter(lambda x: isinstance(x, ClassOptions), values),
                       ClassOptions(ClassOptionsOptions(False, False, False)))

        fields = list(filter(lambda x: isinstance(x, FieldType), values))
        return ClassType(ClassTypeOptions(ctx.getChild(1).getText(), fields, options))


    def visitUnionType(self, ctx: DynabuffersParser.UnionTypeContext):
        self.type_validator.add_available_type(ctx)
        values = []
        for i in range(3, ctx.getChildCount() - 1):
            # TODO: remove workaround
            if str(ctx.getChild(i)) != "{":
                values.append(str(ctx.getChild(i)))

        options = next(filter(lambda x: isinstance(x, UnionOptions), values),
                       UnionOptions(UnionOptionsOptions(False, False, False)))
        return UnionType(UnionTypeOptions(str(ctx.getChild(1)), values, options))

    def visitFieldType(self, ctx: DynabuffersParser.FieldTypeContext):
        values = super().visitFieldType(ctx)
        annotations = list(filter(lambda x: isinstance(x, Annotation), values))
        name = ctx.getChild(len(annotations)).getText()
        datatype = values[len(annotations)]
        options = next(filter(lambda x: isinstance(x, FieldOptions), values), FieldOptions(FieldOptionsOptions(False)))
        defaultVal = next(map(lambda x: x.options.value, filter(lambda x: isinstance(x, Value), values)), None)

        return FieldType(FieldTypeOptions(name, datatype, options, annotations, defaultVal))

    def visitDataType(self, ctx: DynabuffersParser.DataTypeContext):
        if ctx.getText() == "string":
            return StringType(StringTypeOptions("utf-8"))
        if ctx.getText() == "float":
            return FloatType()
        if ctx.getText() == "int":
            return IntType()
        if ctx.getText() == "boolean":
            return BooleanType()
        if ctx.getText() == "long":
            return LongType()
        if ctx.getText() == "byte":
            return ByteType()
        if ctx.getText() == "short":
            return ShortType()
        if ctx.getText() == "map":
            return MapType(MapTypeOptions("utf-8"))
        else:
            self.type_validator.add_used_type(ctx)
            return RefType(RefTypeOptions(ctx.getText()))

    def visitArrayType(self, ctx: DynabuffersParser.ArrayTypeContext):
        datatype = super().visitArrayType(ctx)[0]
        return ArrayType(ArrayTypeOptions(datatype))

    def visitOptionType(self, ctx: DynabuffersParser.OptionTypeContext):
        datatype = super().visitOptionType(ctx)[0]
        return OptionType(OptionTypeOptions(datatype))

    def visitClassOptions(self, ctx: DynabuffersParser.ClassOptionsContext):
        implicit = "implicit" in ctx.getText()
        return ClassOptions(ClassOptionsOptions("primary" in ctx.getText(), "deprecated" in ctx.getText(), implicit))

    def visitUnionOptions(self, ctx: DynabuffersParser.UnionOptionsContext):
        implicit = "implicit" in ctx.getText()
        return UnionOptions(UnionOptionsOptions("primary" in ctx.getText(), "deprecated" in ctx.getText(), implicit))

    def visitFieldOptions(self, ctx: DynabuffersParser.FieldOptionsContext):
        return FieldOptions(FieldOptionsOptions("deprecated" in ctx.getText()))

    def visitAnnotation(self, ctx: DynabuffersParser.AnnotationContext):
        values = super().visitAnnotation(ctx)
        return Annotation(AnnotationOptions(ctx.getChild(1).getText(), values))

    def visitNamespaceType(self, ctx: DynabuffersParser.NamespaceTypeContext):
        _name = ctx.getChild(1).getText()
        _list = list(filter(lambda x: not isinstance(x, NamespaceType), super().visitNamespaceType(ctx)))
        _nestedNamespaces = list(filter(lambda x: isinstance(x, NamespaceType), super().visitNamespaceType(ctx)))
        return NamespaceType(NamespaceTypeOptions(_name, _list), _nestedNamespaces)

    def visitValue(self, ctx: DynabuffersParser.ValueContext):
        if ctx.getText() == "[]":
            return Value(ValueOptions([]))
        if ctx.getText() == "[:]":
            return Value(ValueOptions({}))
        return Value(ValueOptions(ctx.getText()))

    def aggregateResult(self, aggregate, nextResult):
        array = []
        if aggregate is not None:
            array += aggregate
        if nextResult is not None:
            array += [nextResult]
        return array



from dynabuffers.api.ISerializable import ISerializable


class NamespaceTypeOptions:
    def __init__(self, name, list: [ISerializable]):
        self.name = name
        self.list = list


class NamespaceType:

    def __init__(self, options: NamespaceTypeOptions, nestedNamespaces: ['NamespaceType']):
        self.options = options
        self.nestedNamespaces = nestedNamespaces



from dynabuffers.api.ISerializable import ISerializable


class IntType(ISerializable):

    def size(self, value, registry):
        return 4

    def serialize(self, value, buffer, registry):
        buffer.putInt(value if isinstance(value, int) else int(value))

    def deserialize(self, buffer, registry):
        return buffer.getInt()



from typing import List

from dynabuffers.api.ISerializable import ISerializable
from dynabuffers.ast.ClassType import ClassType
from dynabuffers.ast.UnionType import UnionType
from dynabuffers.ast.datatype.OptionType import OptionType
from dynabuffers.ast.datatype.RefType import RefType


class DynabuffersMap(dict):

    def __init__(self, map: dict, tree: List[ISerializable], root: ISerializable):
        super().__init__()
        self.raw_map = map
        self.update(map)
        self.tree = tree
        self.clazz = self.resolve_class(tree, root, map)

    def __getitem__(self, key):
        if not self.is_defined(key):
            raise ValueError("field " + str(key) + " is not defined in " + str(self.clazz.options.name))

        result = self.raw_map[key]
        if result is None and not self.is_optional(key):
            raise ValueError("field " + str(key) + " is not optional in " + str(self.clazz.options.name))

        if result is not None and self.is_ref_type(key):
            ref_type = self.get_ref_type(key)
            if ref_type is not None:
                return DynabuffersMap(result, self.tree, ref_type)

        return result

    def get_ref(self, key: str) -> 'DynabuffersMap':
        if isinstance(self[key], dict):
            return DynabuffersMap(self[key], self.tree, self.clazz)
        raise ValueError("field " + str(key) + " with result " + str(self[key]) + " is not a ref type")

    def get_string(self, key: str) -> str:
        if isinstance(self[key], str):
            return self[key]
        raise ValueError("field " + str(key) + " with result " + str(self[key]) + " is not a string type")

    def get_int(self, key: str) -> int:
        if isinstance(self[key], int):
            return self[key]
        raise ValueError("field " + str(key) + " with result " + str(self[key]) + " is not a int type")

    def get_short(self, key: str) -> int:
        if isinstance(self[key], int):
            return self[key]
        raise ValueError("field " + str(key) + " with result " + str(self[key]) + " is not a short type")

    def get_byte(self, key: str) -> int:
        if isinstance(self[key], int):
            return self[key]
        raise ValueError("field " + str(key) + " with result " + str(self[key]) + " is not a byte type")

    def get_long(self, key: str) -> int:
        if isinstance(self[key], int):
            return self[key]
        raise ValueError("field " + str(key) + " with result " + str(self[key]) + " is not a long type")

    def get_float(self, key: str) -> float:
        if isinstance(self[key], float):
            return self[key]
        raise ValueError("field " + str(key) + " with result " + str(self[key]) + " is not a float type")

    def get_map(self, key: str) -> int:
        if isinstance(self[key], dict):
            return self[key]
        raise ValueError("field " + str(key) + " with result " + str(self[key]) + " is not a map type")

    def get_bytearray(self, key: str) -> bytes:
        if isinstance(self[key], bytes):
            return self[key]
        raise ValueError("field " + str(key) + " with result " + str(self[key]) + " is not a bytearray type")

    def resolve_class(self, tree: List[ISerializable], root: ISerializable, source: dict):
        if isinstance(root, ClassType):
            return root
        if isinstance(root, UnionType):
            if source[":type"] is None:
                return list(filter(lambda x: x.options.name == root.options.values[0], tree))[0]

            type = source[":type"] if isinstance(source[":type"], str) else root.options.values[source[":type"]]
            return list(filter(lambda x: x.options.name == type, tree))[0]
        raise ValueError("cannot resolve class from " + str(root))

    def is_defined(self, key: str) -> bool:
        if key == ":type":
            return True
        return next((x for x in self.clazz.options.fields if x.options.name == key), False)

    def is_optional(self, key: str) -> bool:
        if key == ":type":
            return False
        field = next((x for x in self.clazz.options.fields if x.options.name == key), None)
        return isinstance(field.options.datatype, OptionType) if field is not None else False

    def is_ref_type(self, key: str) -> bool:
        if key == ":type":
            return False
        field = next((x.options.datatype for x in self.clazz.options.fields if x.options.name == key), None)
        return isinstance(field, RefType)

    def get_ref_type(self, key: str) -> ClassType:
        field = next((x for x in self.clazz.options.fields if x.options.name == key), None)
        name = field.options.datatype.options.name
        type = next((x for x in self.tree if x.options.name == name), False)

        source = self.raw_map[key]
        return self.resolve_class(self.tree, type, source) if isinstance(type, UnionType) else type


from typing import List

from dynabuffers.api.ISerializable import ISerializable
from dynabuffers.api.map.DynabuffersMap import DynabuffersMap


class ImplicitDynabuffersMap(DynabuffersMap):

    def __init__(self, map: dict, tree: List[ISerializable], root: ISerializable):
        super().__init__(map, tree, root)

    def get_value(self):
        return self.get("value")


from dynabuffers.api.ISerializable import ISerializable


class StringTypeOptions:
    def __init__(self, charset: str):
        self.charset = charset


class StringType(ISerializable):
    def __init__(self, options: StringTypeOptions):
        self.options = options

    def size(self, value, registry):
        return 4 + len(str(value).encode(self.options.charset))

    def serialize(self, value, buffer, registry):
        encoded = value.encode(self.options.charset)
        buffer.putInt(len(encoded))
        buffer.put(encoded)

    def deserialize(self, buffer, registry):
        length = buffer.getInt()
        array = buffer.get(length)

        return array.decode(self.options.charset)



from antlr4 import *

from dynabuffers.DynabuffersEngine import DynabuffersEngine
from dynabuffers.DynabuffersVisitor import DynabuffersVisitor
from dynabuffers.antlr.DynabuffersLexer import DynabuffersLexer
from dynabuffers.antlr.DynabuffersParser import DynabuffersParser


class Dynabuffers:
    @staticmethod
    def parse(schema) -> DynabuffersEngine:
        lexer = DynabuffersLexer(Dynabuffers.stream(schema))
        stream = CommonTokenStream(lexer)
        parser = DynabuffersParser(stream)

        visitor = DynabuffersVisitor()
        tree = visitor.visit(parser.compilation())

        return DynabuffersEngine(tree)

    @staticmethod
    def stream(obj):
        if isinstance(obj, InputStream):
            return obj
        if isinstance(obj, str):
            return InputStream(str(obj))
        if isinstance(obj, bytes):
            return InputStream(obj.decode("utf-8"))
        if isinstance(obj, bytearray):
            return InputStream(obj.decode("utf-8"))
        if isinstance(obj, DynabuffersResource):
            return obj.get_content_stream()
        raise ValueError(str(obj) + " cannot be converted to input stream")


class DynabuffersResource:

    def __init__(self):
        self.builder = ""

    def append(self, schema):
        if isinstance(schema, str):
            self.builder += str(schema)
        if isinstance(schema, bytes):
            self.builder += schema.decode("utf-8")
        if isinstance(schema, bytearray):
            self.builder += schema.decode("utf-8")
        raise ValueError(str(schema) + " cannot be appended to resource")

    def get_content_stream(self):
        return InputStream(self.builder)



from dynabuffers.api.ISerializable import ISerializable, ByteBuffer
from dynabuffers.ast.ClassType import ClassType
from dynabuffers.ast.structural.UnionOptions import UnionOptions


class UnionTypeOptions():
    def __init__(self, name: str, values: [str], options: UnionOptions):
        self.name = name
        self.values = values
        self.options = options


class UnionType(ISerializable):

    def __init__(self, options):
        self.options = options

    def size(self, value, registry):
        clazz = self.resolve(value, registry)
        return 1 + clazz.size(value, registry)

    def serialize(self, value, buffer: ByteBuffer, registry):
        clazz = self.resolve(value, registry)
        buffer.put(bytes([self.options.values.index(clazz.options.name)]))

        return clazz.serialize(value, buffer, registry)

    def deserialize(self, buffer, registry):
        index = buffer.get()
        clazz = registry.resolve(self.options.values[index])

        map = clazz.deserialize(buffer, registry)
        map[":type"] = index
        return map

    def resolve(self, value, registry) -> ClassType:
        classes = list(map(lambda x: registry.resolve(x), self.options.values))

        for clazz in classes:
            fields1 = sorted(list(map(lambda x: x.options.name, clazz.options.fields)))
            fields2 = sorted(list(value.keys()))

            if ":type" in value and value[":type"] == classes.index(clazz):
                return clazz
            if ":type" in value and value[":type"] in self.options.values \
                    and self.options.values.index(value[":type"]) == classes.index(clazz):
                return clazz
            if fields1 == fields2:
                return clazz

        raise ValueError("union type " + str(value) + " cannot be resolved")



from dynabuffers.api.ISerializable import ISerializable, ByteBuffer


class EnumTypeOptions():
    def __init__(self, name: str, values: [str]):
        self.name = name
        self.values = values


class EnumType(ISerializable):

    def __init__(self, options):
        self.options = options

    def size(self, value, registry):
        return 1 + len(str(value).encode("utf-8"))

    def serialize(self, value, buffer: ByteBuffer, registry):
        array = str(value).encode("utf-8")
        buffer.put(bytes([len(array)]))
        buffer.put(array)

    def deserialize(self, buffer, registry):
        length = buffer.get()
        array = buffer.get(length)

        return array.decode("utf-8")



from dynabuffers.api.ISerializable import ISerializable


class OptionTypeOptions:
    def __init__(self, datatype: ISerializable):
        self.datatype = datatype


class OptionType(ISerializable):

    def __init__(self, options: OptionTypeOptions):
        self.options = options

    def size(self, value, registry):
        return 1 + (0 if value is None else self.options.datatype.size(value, registry))

    def serialize(self, value, buffer, registry):
        buffer.put(bytes([0 if value is None else 1]))
        if value is not None:
            self.options.datatype.serialize(value, buffer, registry)

    def deserialize(self, buffer, registry):
        if buffer.get() == 0:
            return None
        return self.options.datatype.deserialize(buffer, registry)



from dynabuffers.api.ISerializable import ISerializable


class FloatType(ISerializable):

    def size(self, value, registry):
        return 4

    def serialize(self, value, buffer, registry):
        buffer.putFloat(value if isinstance(value, float) else float(value))

    def deserialize(self, buffer, registry):
        return buffer.getFloat()



class ClassOptionsOptions:
    def __init__(self, primary:bool, deprecated:bool, implicit: bool):
        self.primary = primary
        self.deprecated = deprecated
        self.implicit = implicit


class ClassOptions:

    def __init__(self, options: ClassOptionsOptions):
        self.options = options

    def is_primary(self):
        return self.options.primary

    def is_deprecated(self):
        return self.options.deprecated

    def is_implicit(self):
        return self.options.implicit


from dynabuffers.api.IAnnotation import IAnnotation
from dynabuffers.ast.structural import Value


class NotBlank(IAnnotation):

    def __init__(self, args:[Value]):
        self.args = args
        if len(args) != 0:
            raise ValueError("too much annotation values")

    def validate(self, fieldName, value):
        if len(str(value).strip()) == 0:
            raise ValueError("(" + str(fieldName) + ":" + str(value) + ") is blank")



from typing import Union

from antlr4 import ParserRuleContext

from dynabuffers.antlr.DynabuffersParser import DynabuffersParser


class DynabuffersTypeValidator:
    def __init__(self):
        self.visited_data_types = set()
        self.available_data_types = set()

    def validate(self):
        invalid_datatypes = self.visited_data_types - self.available_data_types
        if len(invalid_datatypes) > 0:
            raise ValueError(
                "The datatype(s) {} used in the schema are neither built-in datatypes nor class, union or enum "
                "references.".format(invalid_datatypes))

    def add_available_type(self, ctx: Union[
        DynabuffersParser.ClassTypeContext, DynabuffersParser.UnionTypeContext, DynabuffersParser.EnumTypeContext]):
        name = ctx.getChild(1).getText()
        prefix = self.__get_namespace_prefix(ctx)
        self.available_data_types.add(prefix + name)

    def add_used_type(self, ctx: DynabuffersParser.DataTypeContext):
        name = ctx.getText()
        prefix = self.__get_namespace_prefix(ctx)
        self.visited_data_types.add(prefix + name)

    def __get_namespace_prefix(self, ctx: ParserRuleContext):
        prefix = ""
        while not isinstance(ctx.parentCtx, DynabuffersParser.CompilationContext):
            ctx = ctx.parentCtx
            if isinstance(ctx, DynabuffersParser.NamespaceTypeContext):
                prefix = "{}.{}".format(ctx.getText(), prefix)
        return prefix



from dynabuffers.api.ISerializable import ISerializable


class ByteType(ISerializable):

    def size(self, value, registry):
        return 1

    def serialize(self, value, buffer, registry):
        buffer.put(bytes([value]))

    def deserialize(self, buffer, registry):
        return buffer.get()



from dynabuffers.ast.structural import Value


class AnnotationOptions:
    def __init__(self, name:str, values:[Value]):
        self.name = name
        self.values = values

class Annotation:

    def __init__(self, options: AnnotationOptions):
        self.options = options



from dynabuffers.api.ISerializable import ISerializable


class NoneType(ISerializable):

    def size(self, value, registry):
        return 0

    def serialize(self, value, buffer, registry):
        pass

    def deserialize(self, buffer, registry):
        return None



# Generated from Dynabuffers.g4 by ANTLR 4.7
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .DynabuffersParser import DynabuffersParser
else:
    from DynabuffersParser import DynabuffersParser

# This class defines a complete generic visitor for a parse tree produced by DynabuffersParser.

class DynabuffersVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by DynabuffersParser#compilation.
    def visitCompilation(self, ctx:DynabuffersParser.CompilationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynabuffersParser#enumType.
    def visitEnumType(self, ctx:DynabuffersParser.EnumTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynabuffersParser#classType.
    def visitClassType(self, ctx:DynabuffersParser.ClassTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynabuffersParser#unionType.
    def visitUnionType(self, ctx:DynabuffersParser.UnionTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynabuffersParser#fieldType.
    def visitFieldType(self, ctx:DynabuffersParser.FieldTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynabuffersParser#dataType.
    def visitDataType(self, ctx:DynabuffersParser.DataTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynabuffersParser#arrayType.
    def visitArrayType(self, ctx:DynabuffersParser.ArrayTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynabuffersParser#optionType.
    def visitOptionType(self, ctx:DynabuffersParser.OptionTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynabuffersParser#namespaceType.
    def visitNamespaceType(self, ctx:DynabuffersParser.NamespaceTypeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynabuffersParser#classOptions.
    def visitClassOptions(self, ctx:DynabuffersParser.ClassOptionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynabuffersParser#unionOptions.
    def visitUnionOptions(self, ctx:DynabuffersParser.UnionOptionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynabuffersParser#fieldOptions.
    def visitFieldOptions(self, ctx:DynabuffersParser.FieldOptionsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynabuffersParser#annotation.
    def visitAnnotation(self, ctx:DynabuffersParser.AnnotationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by DynabuffersParser#value.
    def visitValue(self, ctx:DynabuffersParser.ValueContext):
        return self.visitChildren(ctx)



del DynabuffersParser


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dynabuffers',
 'dynabuffers.api',
 'dynabuffers.ast',
 'dynabuffers.ast.datatype',
 'dynabuffers.header']

package_data = \
{'': ['*'],
 'dynabuffers': ['antlr/*'],
 'dynabuffers.api': ['map/*'],
 'dynabuffers.ast': ['annotation/*', 'structural/*']}

install_requires = \
['antlr4-python3-runtime==4.7']

entry_points = \
{'console_scripts': ['build = poetry_scripts:build',
                     'clean = poetry_scripts:clean',
                     'install = poetry_scripts:install',
                     'publish = poetry_scripts:publish',
                     'test = poetry_scripts:test']}

setup_kwargs = {
    'name': 'dynabuffers',
    'version': '2.6.1',
    'description': 'Dynamic cross platform serialization library.',
    'long_description': None,
    'author': 'Leftshift One',
    'author_email': 'contact@leftshift.one',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>3.5.0,<4.0.0',
}


setup(**setup_kwargs)



class FieldOptionsOptions:
    def __init__(self, deprecated:bool):
        self.deprecated = deprecated

class FieldOptions:

    def __init__(self, options: FieldOptionsOptions):
        self.options = options

    def isDeprecated(self):
        return self.options.deprecated



from dynabuffers.api.ISerializable import ISerializable, ByteBuffer
from dynabuffers.ast.structural import FieldOptions


class FieldTypeOptions():
    def __init__(self, name: str, datatype: ISerializable, options: FieldOptions, annotations, defaultVal):
        self.name = name
        self.datatype = datatype
        self.options = options
        self.annotations = annotations
        self.defaultVal = defaultVal


class FieldType(ISerializable):

    def __init__(self, options: FieldTypeOptions):
        self.options = options

    def size(self, value, registry):
        return self.options.datatype.size(value, registry)

    def serialize(self, value, buffer: ByteBuffer, registry):
        if self.options.options:
            registry.addNotification("deprecated field " + self.options.name + " used")

        for annotation in map(lambda x: registry.resolveAnnotation(x.options.name, x.options.values),
                              self.options.annotations):
            annotation.validate(self.options.name, value)

        self.options.datatype.serialize(value, buffer, registry)

    def deserialize(self, buffer, registry):
        return self.options.datatype.deserialize(buffer, registry)



from math import ceil
from typing import List, Union, Any

from dynabuffers.api.ISerializable import ISerializable, ByteBuffer
from dynabuffers.ast.NamespaceType import NamespaceType


class Waypoint:
    def __init__(self, name: str, position: int):
        self._name = name
        self._position = position

    @property
    def name(self):
        return self._name

    @property
    def position(self):
        return self._position


class NamespaceDescription:
    def __init__(self, namespace: NamespaceType, path: List[Waypoint]):
        self._namespace = namespace
        self._path = path

    @property
    def namespace(self):
        return self._namespace

    @property
    def path(self):
        return self._path

    def joined_path(self) -> str:
        return ".".join(list(map(lambda waypoint: waypoint.name, self._path)))


class NamespaceResolver:
    def __init__(self, tree: List[ISerializable]):
        self.tree = tree

    def get_namespace_from_names(self, names: Union[List[str], None]) -> Union[NamespaceDescription, None]:
        if names is None:
            return self._infer_default_namespace()
        r = self.__get_namespace(names, self.__get_namespaces_from_root_level())
        return r

    def get_namespace(self, d: Union[dict, Any]) -> Union[NamespaceDescription, None]:
        namespaces = d.get(':namespace')
        if namespaces is not None:
            return self.get_namespace_from_names(namespaces.split('.'))
        return self._infer_default_namespace()

    def get_namespace_from_serialized(self, bb: ByteBuffer, len_namespace_path: int) -> Union[
        NamespaceDescription, None]:
        """
        Reads the encoded namespace path directly from the given byte buffer
        WARNING: Modifies the provided byte buffer
        :param bb: byte buffer of serialized payload
        :param len_namespace_path: denotes the length of the namespace path (in terms of namespaces)
        :return:
        """
        if len_namespace_path == 0:
            return None

        namespace_bytes = [bb.get() for _ in range(ceil(len_namespace_path / 2))]
        waypoints = []
        # Read waypoints from bytes, 2 waypoints per byte
        for byte in namespace_bytes:
            waypoints.append((byte & 0xF0) >> 4)
            waypoints.append(byte & 0x0F)
        waypoints = waypoints[:len_namespace_path]

        # map the read path to the namespaces within the schema - sorry this is not very elegant
        namespaces = list(map(lambda nd: nd.namespace, self.__get_namespaces_from_root_level()))
        path = []
        for waypoint in waypoints:
            namespace = namespaces[waypoint]
            namespaces = namespace.nestedNamespaces
            path.append(NamespaceDescription(namespace, [Waypoint(namespace.options.name, waypoint)]))

        return NamespaceDescription(path[-1:][0].namespace, list(map(lambda n: n.path[0], path)))

    def __get_namespace(self, names: List[str], namespaces: List[NamespaceDescription]) -> Union[
        NamespaceDescription, None]:
        ns_name = names[0]
        ns = self.__find_namespace_by_name(ns_name, namespaces)
        if len(names) == 1:
            return ns

        nested = [NamespaceDescription(n, ns.path + [Waypoint(n.options.name, idx)]) for idx, n in
                  enumerate(ns.namespace.nestedNamespaces)]
        return self.__get_namespace(names[1:], nested)

    def __get_namespaces_from_root_level(self) -> List[NamespaceDescription]:
        namespace_types = list(filter(lambda x: isinstance(x, NamespaceType), self.tree))
        return [NamespaceDescription(ns, [Waypoint(ns.options.name, idx)]) for idx, ns in enumerate(namespace_types)]

    def __find_namespace_by_name(self, name: str, namespaces: [NamespaceDescription]) -> NamespaceDescription:
        namespace = next((x for x in namespaces if x.namespace.options.name == name), None)
        if namespace is None:
            raise ValueError("no namespace with name " + str(name) + " found")
        return namespace

    def _infer_default_namespace(self) -> Union[NamespaceDescription, None]:
        """
        infers the namespace from the schema - multiple nested namespaces will cause this method to raise an exception.
        :return:
        """

        def _resolve(namespaces: List[NamespaceDescription]) -> Union[NamespaceDescription, None]:
            if len(namespaces) == 0:
                return None
            if len(namespaces) > 1:
                raise ValueError("Could not infer default namespace")
            candidate = namespaces[0]
            if len(candidate.namespace.nestedNamespaces) == 0:
                return candidate
            new_candidates = [NamespaceDescription(ns, candidate.path + [Waypoint(ns.options.name, idx)]) for idx, ns in
                              enumerate(candidate.namespace.nestedNamespaces)]
            return _resolve(new_candidates)

        return _resolve(self.__get_namespaces_from_root_level())



# Generated from Dynabuffers.g4 by ANTLR 4.7
# encoding: utf-8
from antlr4 import *
from io import StringIO
from typing.io import TextIO
import sys

def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3%")
        buf.write("\u009c\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
        buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
        buf.write("\t\16\4\17\t\17\3\2\3\2\3\2\3\2\7\2#\n\2\f\2\16\2&\13")
        buf.write("\2\3\2\3\2\3\3\3\3\3\3\3\3\6\3.\n\3\r\3\16\3/\3\3\3\3")
        buf.write("\3\4\3\4\3\4\5\4\67\n\4\3\4\3\4\6\4;\n\4\r\4\16\4<\3\4")
        buf.write("\3\4\3\5\3\5\3\5\5\5D\n\5\3\5\3\5\6\5H\n\5\r\5\16\5I\3")
        buf.write("\5\3\5\3\6\7\6O\n\6\f\6\16\6R\13\6\3\6\3\6\3\6\3\6\3\6")
        buf.write("\5\6Y\n\6\3\6\5\6\\\n\6\3\6\3\6\5\6`\n\6\3\7\3\7\3\b\3")
        buf.write("\b\3\b\3\b\3\t\3\t\3\t\3\t\3\t\3\t\5\tn\n\t\3\n\3\n\3")
        buf.write("\n\3\n\3\n\3\n\3\n\7\nw\n\n\f\n\16\nz\13\n\3\n\3\n\3\13")
        buf.write("\3\13\6\13\u0080\n\13\r\13\16\13\u0081\3\13\3\13\3\f\3")
        buf.write("\f\6\f\u0088\n\f\r\f\16\f\u0089\3\f\3\f\3\r\3\r\3\r\3")
        buf.write("\r\3\16\3\16\3\16\3\16\3\16\3\16\5\16\u0098\n\16\3\17")
        buf.write("\3\17\3\17\2\2\20\2\4\6\b\n\f\16\20\22\24\26\30\32\34")
        buf.write("\2\6\4\2\n\21\36\36\3\2\36\37\3\2\27\31\4\2\34\36 \"\2")
        buf.write("\u00a3\2$\3\2\2\2\4)\3\2\2\2\6\63\3\2\2\2\b@\3\2\2\2\n")
        buf.write("P\3\2\2\2\fa\3\2\2\2\16c\3\2\2\2\20m\3\2\2\2\22o\3\2\2")
        buf.write("\2\24}\3\2\2\2\26\u0085\3\2\2\2\30\u008d\3\2\2\2\32\u0091")
        buf.write("\3\2\2\2\34\u0099\3\2\2\2\36#\5\4\3\2\37#\5\6\4\2 #\5")
        buf.write("\b\5\2!#\5\22\n\2\"\36\3\2\2\2\"\37\3\2\2\2\" \3\2\2\2")
        buf.write("\"!\3\2\2\2#&\3\2\2\2$\"\3\2\2\2$%\3\2\2\2%\'\3\2\2\2")
        buf.write("&$\3\2\2\2\'(\7\2\2\3(\3\3\2\2\2)*\7\3\2\2*+\7\36\2\2")
        buf.write("+-\7\4\2\2,.\7\36\2\2-,\3\2\2\2./\3\2\2\2/-\3\2\2\2/\60")
        buf.write("\3\2\2\2\60\61\3\2\2\2\61\62\7\5\2\2\62\5\3\2\2\2\63\64")
        buf.write("\7\6\2\2\64\66\7\36\2\2\65\67\5\24\13\2\66\65\3\2\2\2")
        buf.write("\66\67\3\2\2\2\678\3\2\2\28:\7\4\2\29;\5\n\6\2:9\3\2\2")
        buf.write("\2;<\3\2\2\2<:\3\2\2\2<=\3\2\2\2=>\3\2\2\2>?\7\5\2\2?")
        buf.write("\7\3\2\2\2@A\7\7\2\2AC\7\36\2\2BD\5\26\f\2CB\3\2\2\2C")
        buf.write("D\3\2\2\2DE\3\2\2\2EG\7\4\2\2FH\7\36\2\2GF\3\2\2\2HI\3")
        buf.write("\2\2\2IG\3\2\2\2IJ\3\2\2\2JK\3\2\2\2KL\7\5\2\2L\t\3\2")
        buf.write("\2\2MO\5\32\16\2NM\3\2\2\2OR\3\2\2\2PN\3\2\2\2PQ\3\2\2")
        buf.write("\2QS\3\2\2\2RP\3\2\2\2ST\7\36\2\2TX\7\b\2\2UY\5\f\7\2")
        buf.write("VY\5\16\b\2WY\5\20\t\2XU\3\2\2\2XV\3\2\2\2XW\3\2\2\2Y")
        buf.write("[\3\2\2\2Z\\\5\30\r\2[Z\3\2\2\2[\\\3\2\2\2\\_\3\2\2\2")
        buf.write("]^\7\t\2\2^`\5\34\17\2_]\3\2\2\2_`\3\2\2\2`\13\3\2\2\2")
        buf.write("ab\t\2\2\2b\r\3\2\2\2cd\7\22\2\2de\5\f\7\2ef\7\23\2\2")
        buf.write("f\17\3\2\2\2gh\5\f\7\2hi\7\24\2\2in\3\2\2\2jk\5\16\b\2")
        buf.write("kl\7\24\2\2ln\3\2\2\2mg\3\2\2\2mj\3\2\2\2n\21\3\2\2\2")
        buf.write("op\7\25\2\2pq\t\3\2\2qx\7\4\2\2rw\5\4\3\2sw\5\6\4\2tw")
        buf.write("\5\b\5\2uw\5\22\n\2vr\3\2\2\2vs\3\2\2\2vt\3\2\2\2vu\3")
        buf.write("\2\2\2wz\3\2\2\2xv\3\2\2\2xy\3\2\2\2y{\3\2\2\2zx\3\2\2")
        buf.write("\2{|\7\5\2\2|\23\3\2\2\2}\177\7\26\2\2~\u0080\t\4\2\2")
        buf.write("\177~\3\2\2\2\u0080\u0081\3\2\2\2\u0081\177\3\2\2\2\u0081")
        buf.write("\u0082\3\2\2\2\u0082\u0083\3\2\2\2\u0083\u0084\7\32\2")
        buf.write("\2\u0084\25\3\2\2\2\u0085\u0087\7\26\2\2\u0086\u0088\t")
        buf.write("\4\2\2\u0087\u0086\3\2\2\2\u0088\u0089\3\2\2\2\u0089\u0087")
        buf.write("\3\2\2\2\u0089\u008a\3\2\2\2\u008a\u008b\3\2\2\2\u008b")
        buf.write("\u008c\7\32\2\2\u008c\27\3\2\2\2\u008d\u008e\7\26\2\2")
        buf.write("\u008e\u008f\7\30\2\2\u008f\u0090\7\32\2\2\u0090\31\3")
        buf.write("\2\2\2\u0091\u0092\7\33\2\2\u0092\u0097\7\36\2\2\u0093")
        buf.write("\u0094\7\26\2\2\u0094\u0095\5\34\17\2\u0095\u0096\7\32")
        buf.write("\2\2\u0096\u0098\3\2\2\2\u0097\u0093\3\2\2\2\u0097\u0098")
        buf.write("\3\2\2\2\u0098\33\3\2\2\2\u0099\u009a\t\5\2\2\u009a\35")
        buf.write("\3\2\2\2\23\"$/\66<CIPX[_mvx\u0081\u0089\u0097")
        return buf.getvalue()


class DynabuffersParser ( Parser ):

    grammarFileName = "Dynabuffers.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'enum'", "'{'", "'}'", "'class'", "'union'", 
                     "':'", "'='", "'string'", "'short'", "'boolean'", "'byte'", 
                     "'float'", "'long'", "'int'", "'map'", "'['", "']'", 
                     "'?'", "'namespace'", "'('", "'primary'", "'deprecated'", 
                     "'implicit'", "')'", "'@'", "'[]'", "'[:]'", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "';'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "IDENTIFIER", "IDENTIFIER_LITERAL", "STRING", "NUMBER", 
                      "BOOLEAN", "WHITESPACE", "LINE_COMMENT", "SEMICOLON" ]

    RULE_compilation = 0
    RULE_enumType = 1
    RULE_classType = 2
    RULE_unionType = 3
    RULE_fieldType = 4
    RULE_dataType = 5
    RULE_arrayType = 6
    RULE_optionType = 7
    RULE_namespaceType = 8
    RULE_classOptions = 9
    RULE_unionOptions = 10
    RULE_fieldOptions = 11
    RULE_annotation = 12
    RULE_value = 13

    ruleNames =  [ "compilation", "enumType", "classType", "unionType", 
                   "fieldType", "dataType", "arrayType", "optionType", "namespaceType", 
                   "classOptions", "unionOptions", "fieldOptions", "annotation", 
                   "value" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    T__10=11
    T__11=12
    T__12=13
    T__13=14
    T__14=15
    T__15=16
    T__16=17
    T__17=18
    T__18=19
    T__19=20
    T__20=21
    T__21=22
    T__22=23
    T__23=24
    T__24=25
    T__25=26
    T__26=27
    IDENTIFIER=28
    IDENTIFIER_LITERAL=29
    STRING=30
    NUMBER=31
    BOOLEAN=32
    WHITESPACE=33
    LINE_COMMENT=34
    SEMICOLON=35

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    class CompilationContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(DynabuffersParser.EOF, 0)

        def enumType(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynabuffersParser.EnumTypeContext)
            else:
                return self.getTypedRuleContext(DynabuffersParser.EnumTypeContext,i)


        def classType(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynabuffersParser.ClassTypeContext)
            else:
                return self.getTypedRuleContext(DynabuffersParser.ClassTypeContext,i)


        def unionType(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynabuffersParser.UnionTypeContext)
            else:
                return self.getTypedRuleContext(DynabuffersParser.UnionTypeContext,i)


        def namespaceType(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynabuffersParser.NamespaceTypeContext)
            else:
                return self.getTypedRuleContext(DynabuffersParser.NamespaceTypeContext,i)


        def getRuleIndex(self):
            return DynabuffersParser.RULE_compilation

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterCompilation" ):
                listener.enterCompilation(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitCompilation" ):
                listener.exitCompilation(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitCompilation" ):
                return visitor.visitCompilation(self)
            else:
                return visitor.visitChildren(self)




    def compilation(self):

        localctx = DynabuffersParser.CompilationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_compilation)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 34
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << DynabuffersParser.T__0) | (1 << DynabuffersParser.T__3) | (1 << DynabuffersParser.T__4) | (1 << DynabuffersParser.T__18))) != 0):
                self.state = 32
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [DynabuffersParser.T__0]:
                    self.state = 28
                    self.enumType()
                    pass
                elif token in [DynabuffersParser.T__3]:
                    self.state = 29
                    self.classType()
                    pass
                elif token in [DynabuffersParser.T__4]:
                    self.state = 30
                    self.unionType()
                    pass
                elif token in [DynabuffersParser.T__18]:
                    self.state = 31
                    self.namespaceType()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 36
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 37
            self.match(DynabuffersParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class EnumTypeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self, i:int=None):
            if i is None:
                return self.getTokens(DynabuffersParser.IDENTIFIER)
            else:
                return self.getToken(DynabuffersParser.IDENTIFIER, i)

        def getRuleIndex(self):
            return DynabuffersParser.RULE_enumType

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterEnumType" ):
                listener.enterEnumType(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitEnumType" ):
                listener.exitEnumType(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitEnumType" ):
                return visitor.visitEnumType(self)
            else:
                return visitor.visitChildren(self)




    def enumType(self):

        localctx = DynabuffersParser.EnumTypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_enumType)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 39
            self.match(DynabuffersParser.T__0)
            self.state = 40
            self.match(DynabuffersParser.IDENTIFIER)
            self.state = 41
            self.match(DynabuffersParser.T__1)
            self.state = 43 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 42
                self.match(DynabuffersParser.IDENTIFIER)
                self.state = 45 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==DynabuffersParser.IDENTIFIER):
                    break

            self.state = 47
            self.match(DynabuffersParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ClassTypeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(DynabuffersParser.IDENTIFIER, 0)

        def classOptions(self):
            return self.getTypedRuleContext(DynabuffersParser.ClassOptionsContext,0)


        def fieldType(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynabuffersParser.FieldTypeContext)
            else:
                return self.getTypedRuleContext(DynabuffersParser.FieldTypeContext,i)


        def getRuleIndex(self):
            return DynabuffersParser.RULE_classType

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterClassType" ):
                listener.enterClassType(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitClassType" ):
                listener.exitClassType(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitClassType" ):
                return visitor.visitClassType(self)
            else:
                return visitor.visitChildren(self)




    def classType(self):

        localctx = DynabuffersParser.ClassTypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_classType)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 49
            self.match(DynabuffersParser.T__3)
            self.state = 50
            self.match(DynabuffersParser.IDENTIFIER)
            self.state = 52
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==DynabuffersParser.T__19:
                self.state = 51
                self.classOptions()


            self.state = 54
            self.match(DynabuffersParser.T__1)
            self.state = 56 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 55
                self.fieldType()
                self.state = 58 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==DynabuffersParser.T__24 or _la==DynabuffersParser.IDENTIFIER):
                    break

            self.state = 60
            self.match(DynabuffersParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class UnionTypeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self, i:int=None):
            if i is None:
                return self.getTokens(DynabuffersParser.IDENTIFIER)
            else:
                return self.getToken(DynabuffersParser.IDENTIFIER, i)

        def unionOptions(self):
            return self.getTypedRuleContext(DynabuffersParser.UnionOptionsContext,0)


        def getRuleIndex(self):
            return DynabuffersParser.RULE_unionType

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterUnionType" ):
                listener.enterUnionType(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitUnionType" ):
                listener.exitUnionType(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitUnionType" ):
                return visitor.visitUnionType(self)
            else:
                return visitor.visitChildren(self)




    def unionType(self):

        localctx = DynabuffersParser.UnionTypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_unionType)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 62
            self.match(DynabuffersParser.T__4)
            self.state = 63
            self.match(DynabuffersParser.IDENTIFIER)
            self.state = 65
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==DynabuffersParser.T__19:
                self.state = 64
                self.unionOptions()


            self.state = 67
            self.match(DynabuffersParser.T__1)
            self.state = 69 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 68
                self.match(DynabuffersParser.IDENTIFIER)
                self.state = 71 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==DynabuffersParser.IDENTIFIER):
                    break

            self.state = 73
            self.match(DynabuffersParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class FieldTypeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(DynabuffersParser.IDENTIFIER, 0)

        def dataType(self):
            return self.getTypedRuleContext(DynabuffersParser.DataTypeContext,0)


        def arrayType(self):
            return self.getTypedRuleContext(DynabuffersParser.ArrayTypeContext,0)


        def optionType(self):
            return self.getTypedRuleContext(DynabuffersParser.OptionTypeContext,0)


        def annotation(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynabuffersParser.AnnotationContext)
            else:
                return self.getTypedRuleContext(DynabuffersParser.AnnotationContext,i)


        def fieldOptions(self):
            return self.getTypedRuleContext(DynabuffersParser.FieldOptionsContext,0)


        def value(self):
            return self.getTypedRuleContext(DynabuffersParser.ValueContext,0)


        def getRuleIndex(self):
            return DynabuffersParser.RULE_fieldType

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFieldType" ):
                listener.enterFieldType(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFieldType" ):
                listener.exitFieldType(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitFieldType" ):
                return visitor.visitFieldType(self)
            else:
                return visitor.visitChildren(self)




    def fieldType(self):

        localctx = DynabuffersParser.FieldTypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_fieldType)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 78
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==DynabuffersParser.T__24:
                self.state = 75
                self.annotation()
                self.state = 80
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 81
            self.match(DynabuffersParser.IDENTIFIER)
            self.state = 82
            self.match(DynabuffersParser.T__5)
            self.state = 86
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,8,self._ctx)
            if la_ == 1:
                self.state = 83
                self.dataType()
                pass

            elif la_ == 2:
                self.state = 84
                self.arrayType()
                pass

            elif la_ == 3:
                self.state = 85
                self.optionType()
                pass


            self.state = 89
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==DynabuffersParser.T__19:
                self.state = 88
                self.fieldOptions()


            self.state = 93
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==DynabuffersParser.T__6:
                self.state = 91
                self.match(DynabuffersParser.T__6)
                self.state = 92
                self.value()


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class DataTypeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(DynabuffersParser.IDENTIFIER, 0)

        def getRuleIndex(self):
            return DynabuffersParser.RULE_dataType

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterDataType" ):
                listener.enterDataType(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitDataType" ):
                listener.exitDataType(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitDataType" ):
                return visitor.visitDataType(self)
            else:
                return visitor.visitChildren(self)




    def dataType(self):

        localctx = DynabuffersParser.DataTypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_dataType)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 95
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << DynabuffersParser.T__7) | (1 << DynabuffersParser.T__8) | (1 << DynabuffersParser.T__9) | (1 << DynabuffersParser.T__10) | (1 << DynabuffersParser.T__11) | (1 << DynabuffersParser.T__12) | (1 << DynabuffersParser.T__13) | (1 << DynabuffersParser.T__14) | (1 << DynabuffersParser.IDENTIFIER))) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ArrayTypeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def dataType(self):
            return self.getTypedRuleContext(DynabuffersParser.DataTypeContext,0)


        def getRuleIndex(self):
            return DynabuffersParser.RULE_arrayType

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterArrayType" ):
                listener.enterArrayType(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitArrayType" ):
                listener.exitArrayType(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitArrayType" ):
                return visitor.visitArrayType(self)
            else:
                return visitor.visitChildren(self)




    def arrayType(self):

        localctx = DynabuffersParser.ArrayTypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_arrayType)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 97
            self.match(DynabuffersParser.T__15)
            self.state = 98
            self.dataType()
            self.state = 99
            self.match(DynabuffersParser.T__16)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class OptionTypeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def dataType(self):
            return self.getTypedRuleContext(DynabuffersParser.DataTypeContext,0)


        def arrayType(self):
            return self.getTypedRuleContext(DynabuffersParser.ArrayTypeContext,0)


        def getRuleIndex(self):
            return DynabuffersParser.RULE_optionType

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOptionType" ):
                listener.enterOptionType(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOptionType" ):
                listener.exitOptionType(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOptionType" ):
                return visitor.visitOptionType(self)
            else:
                return visitor.visitChildren(self)




    def optionType(self):

        localctx = DynabuffersParser.OptionTypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_optionType)
        try:
            self.state = 107
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [DynabuffersParser.T__7, DynabuffersParser.T__8, DynabuffersParser.T__9, DynabuffersParser.T__10, DynabuffersParser.T__11, DynabuffersParser.T__12, DynabuffersParser.T__13, DynabuffersParser.T__14, DynabuffersParser.IDENTIFIER]:
                self.enterOuterAlt(localctx, 1)
                self.state = 101
                self.dataType()
                self.state = 102
                self.match(DynabuffersParser.T__17)
                pass
            elif token in [DynabuffersParser.T__15]:
                self.enterOuterAlt(localctx, 2)
                self.state = 104
                self.arrayType()
                self.state = 105
                self.match(DynabuffersParser.T__17)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class NamespaceTypeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(DynabuffersParser.IDENTIFIER, 0)

        def IDENTIFIER_LITERAL(self):
            return self.getToken(DynabuffersParser.IDENTIFIER_LITERAL, 0)

        def enumType(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynabuffersParser.EnumTypeContext)
            else:
                return self.getTypedRuleContext(DynabuffersParser.EnumTypeContext,i)


        def classType(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynabuffersParser.ClassTypeContext)
            else:
                return self.getTypedRuleContext(DynabuffersParser.ClassTypeContext,i)


        def unionType(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynabuffersParser.UnionTypeContext)
            else:
                return self.getTypedRuleContext(DynabuffersParser.UnionTypeContext,i)


        def namespaceType(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(DynabuffersParser.NamespaceTypeContext)
            else:
                return self.getTypedRuleContext(DynabuffersParser.NamespaceTypeContext,i)


        def getRuleIndex(self):
            return DynabuffersParser.RULE_namespaceType

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterNamespaceType" ):
                listener.enterNamespaceType(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitNamespaceType" ):
                listener.exitNamespaceType(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitNamespaceType" ):
                return visitor.visitNamespaceType(self)
            else:
                return visitor.visitChildren(self)




    def namespaceType(self):

        localctx = DynabuffersParser.NamespaceTypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_namespaceType)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 109
            self.match(DynabuffersParser.T__18)
            self.state = 110
            _la = self._input.LA(1)
            if not(_la==DynabuffersParser.IDENTIFIER or _la==DynabuffersParser.IDENTIFIER_LITERAL):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 111
            self.match(DynabuffersParser.T__1)
            self.state = 118
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << DynabuffersParser.T__0) | (1 << DynabuffersParser.T__3) | (1 << DynabuffersParser.T__4) | (1 << DynabuffersParser.T__18))) != 0):
                self.state = 116
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [DynabuffersParser.T__0]:
                    self.state = 112
                    self.enumType()
                    pass
                elif token in [DynabuffersParser.T__3]:
                    self.state = 113
                    self.classType()
                    pass
                elif token in [DynabuffersParser.T__4]:
                    self.state = 114
                    self.unionType()
                    pass
                elif token in [DynabuffersParser.T__18]:
                    self.state = 115
                    self.namespaceType()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 120
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 121
            self.match(DynabuffersParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ClassOptionsContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return DynabuffersParser.RULE_classOptions

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterClassOptions" ):
                listener.enterClassOptions(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitClassOptions" ):
                listener.exitClassOptions(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitClassOptions" ):
                return visitor.visitClassOptions(self)
            else:
                return visitor.visitChildren(self)




    def classOptions(self):

        localctx = DynabuffersParser.ClassOptionsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_classOptions)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 123
            self.match(DynabuffersParser.T__19)
            self.state = 125 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 124
                _la = self._input.LA(1)
                if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << DynabuffersParser.T__20) | (1 << DynabuffersParser.T__21) | (1 << DynabuffersParser.T__22))) != 0)):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 127 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << DynabuffersParser.T__20) | (1 << DynabuffersParser.T__21) | (1 << DynabuffersParser.T__22))) != 0)):
                    break

            self.state = 129
            self.match(DynabuffersParser.T__23)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class UnionOptionsContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return DynabuffersParser.RULE_unionOptions

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterUnionOptions" ):
                listener.enterUnionOptions(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitUnionOptions" ):
                listener.exitUnionOptions(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitUnionOptions" ):
                return visitor.visitUnionOptions(self)
            else:
                return visitor.visitChildren(self)




    def unionOptions(self):

        localctx = DynabuffersParser.UnionOptionsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_unionOptions)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 131
            self.match(DynabuffersParser.T__19)
            self.state = 133 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 132
                _la = self._input.LA(1)
                if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << DynabuffersParser.T__20) | (1 << DynabuffersParser.T__21) | (1 << DynabuffersParser.T__22))) != 0)):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 135 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << DynabuffersParser.T__20) | (1 << DynabuffersParser.T__21) | (1 << DynabuffersParser.T__22))) != 0)):
                    break

            self.state = 137
            self.match(DynabuffersParser.T__23)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class FieldOptionsContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return DynabuffersParser.RULE_fieldOptions

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFieldOptions" ):
                listener.enterFieldOptions(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFieldOptions" ):
                listener.exitFieldOptions(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitFieldOptions" ):
                return visitor.visitFieldOptions(self)
            else:
                return visitor.visitChildren(self)




    def fieldOptions(self):

        localctx = DynabuffersParser.FieldOptionsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_fieldOptions)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 139
            self.match(DynabuffersParser.T__19)
            self.state = 140
            self.match(DynabuffersParser.T__21)
            self.state = 141
            self.match(DynabuffersParser.T__23)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class AnnotationContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(DynabuffersParser.IDENTIFIER, 0)

        def value(self):
            return self.getTypedRuleContext(DynabuffersParser.ValueContext,0)


        def getRuleIndex(self):
            return DynabuffersParser.RULE_annotation

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAnnotation" ):
                listener.enterAnnotation(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAnnotation" ):
                listener.exitAnnotation(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAnnotation" ):
                return visitor.visitAnnotation(self)
            else:
                return visitor.visitChildren(self)




    def annotation(self):

        localctx = DynabuffersParser.AnnotationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_annotation)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 143
            self.match(DynabuffersParser.T__24)
            self.state = 144
            self.match(DynabuffersParser.IDENTIFIER)
            self.state = 149
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==DynabuffersParser.T__19:
                self.state = 145
                self.match(DynabuffersParser.T__19)
                self.state = 146
                self.value()
                self.state = 147
                self.match(DynabuffersParser.T__23)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ValueContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self):
            return self.getToken(DynabuffersParser.STRING, 0)

        def NUMBER(self):
            return self.getToken(DynabuffersParser.NUMBER, 0)

        def BOOLEAN(self):
            return self.getToken(DynabuffersParser.BOOLEAN, 0)

        def IDENTIFIER(self):
            return self.getToken(DynabuffersParser.IDENTIFIER, 0)

        def getRuleIndex(self):
            return DynabuffersParser.RULE_value

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterValue" ):
                listener.enterValue(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitValue" ):
                listener.exitValue(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitValue" ):
                return visitor.visitValue(self)
            else:
                return visitor.visitChildren(self)




    def value(self):

        localctx = DynabuffersParser.ValueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_value)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 151
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << DynabuffersParser.T__25) | (1 << DynabuffersParser.T__26) | (1 << DynabuffersParser.IDENTIFIER) | (1 << DynabuffersParser.STRING) | (1 << DynabuffersParser.NUMBER) | (1 << DynabuffersParser.BOOLEAN))) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx





